part of my_hk_reader;

class MyHkReader {
  static const MethodChannel _channel = const MethodChannel('my_hk_reader');

  static Future<String?> getPlatformVersion() async {
    return await _channel.invokeMethod("getPlatformVersion");
  }

  /// iOS isn't completely supported by HealthKit, false means no, true means user has approved or declined permissions.
  /// In case user has declined permissions read will just return empty list for declined data types.
  static Future<bool?> hasPermissions(List<DataType> types) async {
    return await _channel.invokeMethod('hasPermissions', {
      "types": types.map((type) => _dataTypeToString(type)).toList(),
    });
  }

  /// If you're using more than one DataType it's advised to call requestPermissions with all the data types once,
  /// otherwise iOS HealthKit will ask to approve every permission one by one in separate screens.
  ///
  /// `await FitKit.requestPermissions(DataType.values)`
  static Future<bool?> requestPermissions(List<DataType> types) async {
    return await _channel.invokeMethod('requestPermissions', {
      "types": types.map((type) => _dataTypeToString(type)).toList(),
    });
  }

  /// iOS isn't supported by HealthKit, method does nothing.
  static Future<void> revokePermissions() async {
    return await _channel.invokeMethod('revokePermissions');
  }

  /// #### It's not advised to call `await FitKit.read(dataType)` without any extra parameters. This can lead to FAILED BINDER TRANSACTION on Android devices because of the data batch size being too large.
  static Future<List<FitData>> read(
    DataType type, {
    DateTime? dateFrom,
    DateTime? dateTo,
    int? limit,
  }) async {
    return await _channel
        .invokeListMethod('read', {
          "type": _dataTypeToString(type),
          "date_from": dateFrom?.millisecondsSinceEpoch ?? 1,
          "date_to": (dateTo ?? DateTime.now()).millisecondsSinceEpoch,
          "limit": limit,
        })
        .then(
          (response) =>
              response!.map((item) => FitData.fromJson(item)).toList(),
        )
        .catchError(
          (_) => throw UnsupportedException(type),
          test: (e) {
            if (e is PlatformException) return e.code == 'unsupported';
            return false;
          },
        );
  }

  static Future<List<dynamic>> readSteps(DataType type,
      {required DateTime dateFrom, required DateTime dateTo}) async {
    List<DateTime> targets = makeTargetDates(dateFrom, dateTo);

    List<dynamic> results = [];

    for (var i = 0; i < targets.length; i++) {
      Map<String, dynamic> result;
      List<DateTime> range = makeDateRange(targets[i]);
      try {
        var response = await readSpecifiedDateSteps(type, range[0], range[1]);
        result = {
          "dateFrom": range[0],
          "dateTo": range[1],
          "value": response[0],
          "userEntered": false,
        };
      } catch (ex) {
        result = {
          "dateFrom": range[0],
          "dateTo": range[1],
          "value": 0.0,
          "userEntered": false
        };
      }
      results.add(result);
    }

    return results;
  }

  static Future<List<dynamic>> readDistances(DataType type,
      {required DateTime dateFrom, required DateTime dateTo}) async {
    List<DateTime> targets = makeTargetDates(dateFrom, dateTo);

    List<dynamic> results = [];

    for (var i = 0; i < targets.length; i++) {
      Map<String, dynamic> result;
      List<DateTime> range = makeDateRange(targets[i]);
      try {
        var response =
            await readSpecifiedDateDistances(type, range[0], range[1]);
        result = {
          "dateFrom": range[0],
          "dateTo": range[1],
          "value": response[0],
          "userEntered": false,
        };
      } catch (ex) {
        result = {
          "dateFrom": range[0],
          "dateTo": range[1],
          "value": 0.0,
          "userEntered": false
        };
      }
      results.add(result);
    }

    return results;
  }

  static List<DateTime> makeTargetDates(DateTime from, DateTime to) {
    List<DateTime> result = [];
    var tmp = from;

    while (new DateTime(tmp.year, tmp.month, tmp.day)
            .isBefore(new DateTime(to.year, to.month, to.day)) ||
        isSameDay(new DateTime(tmp.year, tmp.month, tmp.day),
            new DateTime(to.year, to.month, to.day))) {
      result.add(new DateTime(tmp.year, tmp.month, tmp.day));
      tmp = new DateTime(tmp.year, tmp.month, tmp.day).add(Duration(days: 1));
    }
    return result;
  }

  static bool isSameDay(DateTime date1, DateTime date2) {
    var diff = ((date1.difference(date2) -
                    Duration(hours: date1.hour) +
                    Duration(hours: date2.hour))
                .inHours /
            24)
        .round();
    return diff == 0;
  }

  static List<DateTime> makeDateRange(DateTime date) {
    DateTime start = date;
    DateTime end = date.add(Duration(days: 1)).subtract(Duration(seconds: 1));
    return [start, end];
  }

  static Future<List<dynamic>> readSpecifiedDateSteps(
      DataType type, DateTime dateFrom, DateTime dateTo) async {
    return await (_channel
        .invokeMethod('readSteps', {
          "type": _dataTypeToString(type),
          "date_from": dateFrom.millisecondsSinceEpoch,
          "date_to": (dateTo).millisecondsSinceEpoch,
        })
        .then((value) => value)
        .catchError((_) => throw UnsupportedException(type), test: (e) {
          if (e is PlatformException) return e.code == 'unsupported';
          return false;
        }) as FutureOr<List<dynamic>>);
  }

  static Future<List<dynamic>> readSpecifiedDateDistances(
      DataType type, DateTime dateFrom, DateTime dateTo) async {
    return await (_channel
        .invokeMethod('readDistances', {
          "type": _dataTypeToString(type),
          "date_from": dateFrom.millisecondsSinceEpoch,
          "date_to": (dateTo).millisecondsSinceEpoch,
        })
        .then((value) => value)
        .catchError((_) => throw UnsupportedException(type), test: (e) {
          if (e is PlatformException) return e.code == 'unsupported';
          return false;
        }) as FutureOr<List<dynamic>>);
  }

  // static Future<FitData> readLast(DataType type) async {
  //   return await read(type, limit: 1)
  //       .then((results) => results.isEmpty ? null : results[0]);
  // }

  static String _dataTypeToString(DataType type) {
    switch (type) {
      case DataType.HEART_RATE:
        return "heart_rate";
      case DataType.STEP_COUNT:
        return "step_count";
      case DataType.HEIGHT:
        return "height";
      case DataType.WEIGHT:
        return "weight";
      case DataType.DISTANCE:
        return "distance";
      case DataType.ENERGY:
        return "energy";
      case DataType.WATER:
        return "water";
      case DataType.SLEEP:
        return "sleep";
      case DataType.STAND_TIME:
        return "stand_time";
      case DataType.EXERCISE_TIME:
        return "exercise_time";
    }
  }
}

enum DataType {
  HEART_RATE,
  STEP_COUNT,
  HEIGHT,
  WEIGHT,
  DISTANCE,
  ENERGY,
  WATER,
  SLEEP,
  STAND_TIME,
  EXERCISE_TIME
}

class UnsupportedException implements Exception {
  final DataType dataType;
  UnsupportedException(this.dataType);

  @override
  String toString() => 'UnsupportedException: dataType $dataType not supported';
}
